package ca.recrown.islandsurvivalcraft.world;

import java.util.Random;

import org.bukkit.util.noise.SimplexOctaveGenerator;

import ca.recrown.islandsurvivalcraft.utilities.caching.Cache;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.utilities.datatypes.Reference;
import ca.recrown.islandsurvivalcraft.utilities.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.utilities.pathfinding.DepthFirstSearch;

public class IslandWorldMap {
    private final float SHALLOW_PORTION = 0.06f;
    private final float TRANSITION_DEPTH_PORTION = 0.1f;
    private final float DEEP_OCEAN_PORTION = 0.5f;
    private final float HILL_PORTION = 0.4f;
    private final float SHORE_PORTION = 0.095f;

    private final Cache<Point2, Double> blockValueCache;
    private final SimplexOctaveGenerator noiseGenerator;
    public final islandValidator islandValidator = new islandValidator();
    private final int NOISE_OCTAVES = 4;
    private final float ISLAND_PERCENT = 0.36f;
    private final double NOISE_FREQ = 1.78D;
    private final double NOISE_AMP = 0.47D;
    private final double SCALE = 0.005D;

    public IslandWorldMap(Random random) {
        this.noiseGenerator = new SimplexOctaveGenerator(random, NOISE_OCTAVES);
        noiseGenerator.setScale(SCALE);
        this.blockValueCache = new Cache<>(256000);
    }
    
    /**
     * Considers this land if the world block value is greater than or equal to 0.
     * @param worldX the X coordinate of location in world.
     * @param worldZ the Z coordinate of location in world.
     * @return
     */
    public boolean isLand(int worldX, int worldZ) {
        if (getWorldValue(worldX, worldZ) >= 0) {
            return true;
        }
        return false;
    }

    /**
     * Calculates if this is the edge of the island by checking if it's surrounding blocks are part of this island.
     * @param worldX The world X coordinate.
     * @param worldZ The world Z coordinate.
     * @return true if it is the edge of the island and false if it isn't.
     */
    public boolean isEdgeOfIsland(int worldX, int worldZ) {
        return isIsland(worldX, worldZ) &&
        !(isIsland(worldX + 1, worldZ) && 
        isIsland(worldX - 1, worldZ) && 
        isIsland(worldX , worldZ + 1) && 
        isIsland(worldX, worldZ - 1));
    }

    /**
     * Considers these coordinates to be part of an island if:
     * It is land, or if it is a shallow portion of land.
     * @param worldX the X coordinate of location in world.
     * @param worldZ the Y coordinate of location in world.
     * @return true if these coordinates represent an island.
     */
    public boolean isIsland(int worldX, int worldZ) {
        return isLand(worldX, worldZ) || isShallowPortion(worldX, worldZ);
    }

    /**
     * Checks coordinates for if this location is a hill.
     * It is considered a hill when its world value is greater than the HILL_PORTION.
     * @param worldX
     * @param worldZ
     * @return
     */
    public boolean isHill(int worldX, int worldZ) {
        return getWorldValue(worldX, worldZ) >= HILL_PORTION;
    }

    public boolean isShore(int worldX, int worldZ) {
        return isLand(worldX, worldZ) && getWorldValue(worldX, worldZ) <= SHORE_PORTION;
    }

    public boolean isTransitional(int worldX, int worldZ) {
        return !isIsland(worldX, worldZ) && getWorldValue(worldX, worldZ) >= -TRANSITION_DEPTH_PORTION;
    }

    public boolean isDeep(int worldX, int worldZ) {
        return getWorldValue(worldX, worldZ) <= -DEEP_OCEAN_PORTION;
    }

    /**
     * This block is considered a shallow portion of the island:
     * The block height is less than the sea level,
     * and that the block height is greater than the shallow portion.
     * If it is land, it is not shallow.
     * @param worldX the X coordinate of location in world.
     * @param worldZ the Z coordinate of location in world.
     * @return true if it is considered the shallow portion.
     */
    public boolean isShallowPortion(int worldX, int worldZ) {
        if (!isLand(worldX, worldZ) && getWorldValue(worldX, worldZ) >= -SHALLOW_PORTION) {
            return true;
        }
        return false;
    }


    /**
     * World block value will be 0 or positive if it is a part of an island.
     * If less than, it is considered under the sea.
     * Does not factor in a shallow depth.
     * The value is normalized to [-1, 1].
     * @param worldX the x world coordinate to obtain this value for.
     * @param worldZ the z world coordinate to obtain this value for.
     * @return a value representing the island at the given point.
     */
    public double getWorldValue(int worldX, int worldZ) {
        Point2 p = new Point2(worldX, worldZ);

        Double res = blockValueCache.get(p);
        if (res == null) {
            double shift = 1f - 2f * ISLAND_PERCENT;
            double rawNoise = noiseGenerator.noise(worldX, worldZ, NOISE_FREQ, NOISE_AMP, true) - shift;
            double maxNeg = -1 - shift;
            double maxPos = 1 - shift;
            
            if (rawNoise < 0) {
                res = - rawNoise / maxNeg;
            } else {
                res = rawNoise / maxPos;
            }
            blockValueCache.set(p, res);
            return res;
        }

        return res;
    }

    /**
     * Determines of the two columns are part of the same island.
     * If both points are connected consistently by island, than both are considered part of the same island.
     * @param firstColumn The column of the first island.
     * @param secondColumn The column of the second island.
     * @return If the two islands are connected.
     */
    public boolean isSameIsland(Point2 firstColumn, Point2 secondColumn) {
        DepthFirstSearch dfs = new DepthFirstSearch(firstColumn, secondColumn);
        boolean same = dfs.buildToGoal(islandValidator, null);
        return same;
    }

    /**
     * Find the island's origin block. Will call the coordinate target validatable to end early if nessecary.
     * @param worldX The x coordinate of the island block in question.
     * @param worldZ The y coordinate of the island block in question.
     * @param targetValidator The coordinate target validator to use. Optional and is null.
     * @return The island origin point. Will be null if the passed target target validator stops search prematurely.
     */
    public Point2 findIslandOrigin(int worldX, int worldZ, CoordinateValidatable targetValidator) {
        if (!isIsland(worldX, worldZ)) throw new IllegalArgumentException("The given coordinates are not part is an island.");
        final Point2 goal = new Point2(0, 0);
        DepthFirstSearch dfs = new DepthFirstSearch(new Point2(worldX, worldZ), goal);
        Reference<Point2> closest = new Reference<>();
        dfs.buildToGoal(islandValidator, new CoordinateValidatable() {
            Double closestDistance = null;
            @Override
            public boolean validate(Point2 point) {
                if (targetValidator != null && targetValidator.validate(point)) {
                    closest.value = null;
                    return true;
                }
                double currentDistance = point.distance(goal);
                if (closestDistance == null || (currentDistance) < closestDistance) {
                    closestDistance = currentDistance;
                    closest.value = point;
                }
                return false;
            }
        });

        return closest.value;
    }

    public Point2 findIslandOrigin(int worldX, int worldZ) {
        return findIslandOrigin(worldX, worldZ, null);
    }

    public int calculateIslandHash(int worldX, int worldZ) {
        return findIslandOrigin(worldX, worldZ).hashCode();
    }

    private class islandValidator implements CoordinateValidatable {

        @Override
        public boolean validate(Point2 point) {
            return isIsland(point.x, point.y);
        }

    }
}